#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <fcntl.h>

#include <io.h>
#include "bit_byte.h"

bit_byte outbit;
bit_byte inbit;
unsigned int extraout;
unsigned int put1flag;

unsigned int extrain;
unsigned int get1flag;
unsigned int pwt[256],pnd[256];
unsigned int ptv[256],pts[256];
unsigned int empty1,empty2;



//  please address any bugs discovered in this code
//  to the author: karl malbrain, karl_m@acm.org
/* heavily modifed by David Scott to make bijective
   errors where found and corrected but use at your
   own risk */

//  link bit-level I/O

void arc_put1 (unsigned bit);
void arc_put2 (unsigned bit);
unsigned arc_get1 ();
unsigned arc_get2 ();

//  This code is adapted from Professor Vitter's
//  article, Design and Analysis of Dynamic Huffman Codes,
//  which appeared in JACM October 1987

//  A design trade-off has been made to simplify the
//  code:  a node's block is determined dynamically.

//  The dynamic huffman table weight ranking
//  is maintained per Professor Vitter's
//  invariants for algorithm FGK:

//  1) leaves preceed internal nodes of the
//  same weight in a non-decreasing
//  ranking of weights;

//  2) equal weight internal nodes are kept
//  ordered by their tree positions.

//  weights are incremented by 2:
//  leaves always have even weight values;
//  internal nodes always have odd values.

//  tree nodes (HuffSize is the alphabet size):
//  nodes >= HuffSize are internal nodes
//  nodes < HuffSize are leaf nodes

//  node 2 * HuffSize - 1 is always the tree root;
//  node HuffEsc is always the escape node;

//  the tree is initialized by creating an
//  escape node as the root.

//  each new leaf symbol is paired with a new escape
//  node into the previous escape node, until the
//  last symbol, which is incorporated into the
//  tree into the position of the escape node.

//  The orphaned escape node is left with its
//  up pointer set to zero, and its next
//  pointing at the last symbol.

//  table size: 2 * HuffSize

//  huff_init(alphabet_size)
//  huff_encode(next_symbol)
//  next_symbol = huff_decode()

//  huff_scale(by_bits) -- scale weights and rebalance tree

typedef struct {
    unsigned parity,  // 1:right link, 0:left link
        up,           // the next node up the tree
        prev,         // next <= weight node
        next,         // next >= weight node
        dnl,          // next node down-left
        dnr,          // next node down-right
        weight;       // node weight
} HTable;

typedef struct {
    unsigned esc,     // the current tree height
        size;         // the size of the alphabet
    HTable table[1];  // the coding table starts here
} HCoder;

HCoder *huff_init (unsigned size)
{
unsigned esc = 2 * size - 1;
HCoder *huff;

  //  create the initial escape node
  //  at the tree root

  huff = (HCoder *) malloc (esc * sizeof(HTable) + sizeof(HCoder));
  memset (huff->table, 0, 2 * size * sizeof(HTable));
  huff->table[esc].weight = 1;
  huff->table[esc].next = esc;
  huff->size = size;
  huff->esc = esc;
  return huff;
}

// split escape node to incorporate new symbol

void huff_split (HCoder *huff, unsigned node)
{
unsigned next, pair;

    //  the final symbol takes the escape node's left child
    //  tree position, and the orphan escape node links to
    //  the first symbol in the rank list (this last symbol)

    if( huff->esc == huff->size ) {
        next = huff->table[huff->esc].next;
        huff->table[huff->esc + 1].dnl = node;
        huff->table[huff->esc].next = node;
        huff->table[huff->esc].up = 0;
        huff->table[next].prev = node;

        huff->table[node].up = huff->esc + 1;
        huff->table[node].prev = huff->esc;
        huff->table[node].next = next;
        huff->table[node].parity = 0;
        huff->table[node].weight = 0;
        return;
    }

    //  other symbols initialize a new escape node.
    //  the old escape node is promoted to parent
    //  the new escape node and the new symbol.

    //  HuffEsc is always the current escape node;
    //  promote the old escape node.

    pair = huff->esc--;
    huff->table[pair].dnl = huff->esc;
    huff->table[pair].prev = node;
    huff->table[pair].dnr = node;

    //  create a new escape node.

    huff->table[huff->esc].next = node;
    huff->table[huff->esc].weight = 1;
    huff->table[huff->esc].parity = 0;
    huff->table[huff->esc].up = pair;

    //  initialize the new symbol to the
    //  right of the new escape node
    //  and beneath the promoted
    //  pair node at pair

    huff->table[node].prev = huff->esc;
    huff->table[node].next = pair;
    huff->table[node].parity = 1;
    huff->table[node].weight = 0;
    huff->table[node].up = pair;
}

//  send the bits for an escaped symbol

void huff_sendid (HCoder *huff, unsigned node)
{
/* this routine modifed by David Scott since
it was wrong it often used more bits that
required for new symbol */

unsigned empty = 0, max;

unsigned dsx,dsy,dsz = 0;

    //  count the number of empty nodes
    //  before the symbol in the table

//    while( node-- )
//      if( !huff->table[node].weight )
//        empty++;


    //  send LSB of this count first, using
    //  as many bits as are required for
    //  the maximum possible count
/* first error you sometimes need one less bit here
   the original code off by one bit sometimes so fixed */
    max = huff->esc - huff->size;
    dsx = pnd[max];
    empty = pwt[node];
    pwt[dsx] = empty;
    pnd[empty] = dsx;
    empty1 = empty;
    empty += max + 1 - empty2;
    if ( empty > max ) empty -= max+1;
    empty2 = empty1;
    dsx = max + max;
    for ( dsy = 1; dsx >>= 1; dsy <<= 1);
    dsx = dsy - max - 1;
    dsz = dsy >> 1;
    if ( empty < dsx )  max >>= 1;
    else if ( empty >= dsz ) empty += dsx;
    if( max )
      do  arc_put1 (empty & 1), empty >>= 1;
      while( max >>= 1 );
}

//  slide node to right over all leaves of equal weight
//  return node that follows

unsigned huff_slide (HCoder *huff, unsigned node, unsigned idx)
{
unsigned prev = huff->table[node].prev;
unsigned next = huff->table[idx].next;
unsigned slide, up, parity;

    // find rightmost possible leaf to exchange with

    if( idx < huff->size )
      while( huff->table[next].weight == huff->table[idx].weight )
          idx = next, next = huff->table[idx].next;

    // swap the tree positions
    // of node and idx.

    up = huff->table[idx].up;

    if( parity = huff->table[idx].parity )
        huff->table[up].dnr = node;
    else
        huff->table[up].dnl = node;

    huff->table[idx].parity = huff->table[node].parity;
    huff->table[idx].up = huff->table[node].up;
    huff->table[node].parity = parity;
    huff->table[node].up = up;

    up = huff->table[idx].up;

    if( huff->table[idx].parity )
        huff->table[up].dnr = idx;
    else
        huff->table[up].dnl = idx;

    //  exchange the ranking positions
    //  of node and idx.

    slide = huff->table[idx].prev;

    //  simple swap??

    if( slide == node ) {
        huff->table[node].prev = idx;
        huff->table[idx].next = node;
    } else {
        huff->table[node].prev = slide;
        huff->table[slide].next = node;
        slide = huff->table[node].next;
        huff->table[slide].prev = idx;
        huff->table[idx].next = slide;
    }

    // we never swap the escape node nor
    // the root node -- they will always
    // have the smallest and largest
    // weights in the tree

    huff->table[prev].next = idx;
    huff->table[next].prev = node;

    huff->table[idx].prev = prev;
    return huff->table[node].next = next;
}

//  increment node weights and re balance the tree.

void huff_increment (HCoder *huff, unsigned node)
{
unsigned next, up = huff->table[node].up;

  //  obviate swapping a parent with its child:
  //    increment the leaf and proceed
  //    directly to its parent.

  if( huff->table[node].next == up )
      huff->table[node].weight += 2;
  else
      up = node;

  //  slide right and go up until reaching the root

  while( node = up, huff->table[node].up ) {
    next = huff->table[node].next;

    //  promote the node to group leader
    //  position by sliding right over
    //  any equal weight nodes

    while( huff->table[node].weight == huff->table[next].weight )
        next = huff_slide (huff, node, next);

    //  increase the weight of the node

    huff->table[node].weight += 2;

    //  internal nodes go up from this
    //  initial group leader position

    if( node > huff->size )
        up = huff->table[node].up;

    //  slide incremented node over smaller weights to its right

    while( huff->table[node].weight > huff->table[next].weight )
        next = huff_slide (huff, node, next);

    //  symbol nodes slide over first,
    //  then go up from their
    //  final positions

    if( node < huff->size )
        up = huff->table[node].up;
  }

  //  increase the root's weight

  huff->table[node].weight += 2;
}

//  scale all weights and rebalance the tree

//  zero weight nodes are removed from the tree
//  by sliding them out the left of the rank list

void huff_scale (HCoder *huff, unsigned bits)
{
unsigned node, next, prev;
unsigned weight;

  //  start from the node following the escape node
  //  proceed across the nodes in weight order
  //  scaling them by the value of bits

  //  clear the current escape node weight to serve
  //  as a stopper for newly zeroed weight leaves
return; /***temp***/

  huff->table[huff->esc].weight = 0;
  next = huff->table[huff->esc].next;

  while( node = next, huff->table[node].up ) {
    next = huff->table[node].next;

    //  recompute the weight of internal nodes
    //  and slide out unused ones to the left
    //  with a zero weight

    if( node > huff->size ) {
      if( weight = huff->table[huff->table[node].dnr].weight & ~1 )
        weight += huff->table[huff->table[node].dnl].weight | 1;

    //  remove zero weight leaves by incrementing HuffEsc
    //  or, if table is full, re-splice escape node into tree
    //  without the zero weight leaf

    } else if( !(weight = huff->table[node].weight >> bits & ~1) )
      if( huff->table[huff->esc].up )
        huff->esc++;
      else {
        huff->table[huff->esc + 1].dnl = huff->esc;
        huff->table[huff->esc].up = huff->esc + 1;
        huff->table[next].prev = huff->esc;
        huff->table[huff->esc].next = next;
      }

    // slide the scaled node back over any
    // previous nodes with larger weights

    huff->table[node].weight = weight;
    prev = huff->table[node].prev;

    while( weight < huff->table[prev].weight )
        huff_slide (huff, prev, node), prev = huff->table[node].prev;
  }

  //  reset the (possibly new) escape node's weight

  huff->table[huff->esc].weight = 1;
}

//  encode the next symbol

void huff_encode (HCoder *huff, unsigned node)
{
unsigned emit = 1, bit;
unsigned up, idx;

    //  transform illegal symbols to zero

    if( node >= huff->size )
        node = 0;

    //  for a new symbol, direct the receiver to the escape node

    if( huff->table[node].weight )
        idx = node;
      else
        idx = huff->esc;

    //  accumulate the code bits by
    //  working up the tree from
    //  the node to the root

    while( up = huff->table[idx].up )
        emit <<= 1, emit |= huff->table[idx].parity, idx = up;

    //  send the code, root selector bit first

    while( bit = emit & 1, emit >>= 1 )
        arc_put1 (bit);

    //  send identification and incorporate
    //  new symbols into the tree

    if( !huff->table[node].weight )
        huff_sendid(huff, node), huff_split(huff, node);

    //  adjust and re-balance the tree

    huff_increment (huff, node);
}

//  read the identification bits
//  for an escaped symbol

unsigned huff_readid (HCoder *huff)
{
unsigned empty = 0, bit = 1, node, max;
unsigned dsx,dsy,dsz = 0;

    //  receive the code, LSB first, reading
    //  only the number of bits necessary to
    //  transmit the maximum possible value
/* again this is in error so fixing since it
often reads more bits that necessary */

    max = huff->esc - huff->size;
    node = max;
    if( max >>= 1) /* do one less since may be all needed */
      do empty |= arc_get1 () ? bit : 0, bit <<= 1;
      while( max >>= 1 );
    max = huff->esc - huff->size;
    dsx = max + max;
    for ( dsy = 1; dsx >>= 1; dsy <<= 1);
    dsx = dsy - max - 1;
    dsz = dsy >> 1;
    if ( empty >= dsx && max != 0 )  /* get one more bit */
      if( arc_get1() == 1) empty += dsz-dsx; 
    dsz = max;
    empty += empty2;
    if ( empty > max ) empty -= max+1;
    empty2 = empty;
    max = pnd[node]; 
    node = pnd[empty];
    pwt[max] = empty;
    pnd[empty] = max;

    //  the count is of zero weight
    //  symbols in the table before
    //  the zero weight new symbol

//    for( node = 0; node < huff->size; node++ )
//      if( !huff->table[node].weight )
//        if( !empty-- )

            return node;

    //  oops!  our count was too big!
/* adding a check here may do it on last but should no where else */
   printf ( " **should never get here** \n");
    return 0;
}

//  decode the next symbol

unsigned huff_decode (HCoder *huff)
{
unsigned node = 2 * huff->size - 1;

    //  work down the tree from the root
    //  until reaching either a leaf
    //     or the escape node

    while( node > huff->esc )
        if( arc_get1 () )
            node = huff->table[node].dnr;
        else
            node = huff->table[node].dnl;

    //  sent to the escape node???

    if( node == huff->esc )
        node = huff_readid (huff), huff_split (huff, node);

    //  increment weights and rebalance
    //  the coding tree

    huff_increment (huff, node);
    return node;
}
#define HUFFSTANDALONE 1

#ifdef HUFFSTANDALONE


FILE *In, *Out, *Cond;
unsigned char ArcBit = 0;
int ArcChar = 0;

int main (int argc, char **argv)
{
int mode, size, symbol;
HCoder *huff;

    mode = argv[1][0], argv[1]++;
    if( argc != 5 || !( mode == 'd' || mode == 'c') )
    {
        printf ("for compression: %s c infile outfile conditonfile\nnn -- alphabet size\ninfile -- source file\noutfile -- output file\n", argv[0]);
        printf ("for decompression: %s d infile outfile condtionfile\nnn -- alphabet size\ninfile -- source file\noutfile -- output file\n", argv[0]);
        printf (" **condtion** as symbol first appers assigned a slot \n");
        printf (" each addtion appearance cause it to build huffman tree \n");
                   
        return 1;
    }

// add scaling later if one wishes

    if( argc > 3 )
      if( !(Out = fopen (argv[3], "wb")) )
        return 1;
    if ( !(Cond = fopen (argv[4], "rb")) )
        return 1;

    //  literal text
/* I dropped this for bijective adaptive huffman */

    //  alphabet fill
/* I dropped this for bijective adaptive huffman */


    if( argc > 2 )
      if( !(In = fopen (argv[2], "rb")) )
        return 1;

/* read Cond file */
   size = 0;
      for ( symbol = 256; symbol != 0;){
        symbol--;
        ptv[symbol] = 777;
        pts[symbol] = 777;
      }
    while( 1 ) {
      if( (symbol = getc(Cond) ) == EOF ) break;
      if ( ptv[symbol] == 777) {
         ptv[symbol] = size;
         pts[size] = symbol;
         size++;
      }
    }
if ( size > 256 || size < 3 ) 
  printf(" size % has to be in range 3 to 256 try again \n"),exit(2);
    rewind(Cond);
      for ( symbol = 256; symbol != 0;){
        symbol--;
        pwt[symbol] = symbol;
        pnd[symbol] = symbol;
      }
printf(" number of symbols allowed is %u \n",size);
    huff = huff_init (size);
    size = 0;
    put1flag = 777;
    while( 1 ) {
      if( (symbol = getc(Cond) ) == EOF ) break;
      symbol = ptv[symbol];
      size ++;
      huff_encode(huff, symbol);
   }
printf(" the number of symbol values used to make tree %u \n",size);
    //  decompression

    if( mode == 'd' ) {
/* use same way as for compression */
         
      inbit.ir(In);
      extrain = 0;
      get1flag = 0;
     empty1 = 0;
     empty2 = 0;
      while( 1 ) {
        arc_get2();
        symbol = huff_decode( huff);
        symbol = pts[symbol];
        if (extrain == 300 )break;
        putc (symbol, Out);
        if (extrain == 400)break;
      }
      return 0;
    }

    // compression

    outbit.iw(Out);
    extraout = 0;
    put1flag = 0;
    empty1 = 0;
    empty2 = 0;
    while( 1 ) {
      arc_put2(1);
      if( (symbol = getc(In) ) == EOF ) break;
      symbol = ptv[symbol];
      if ( symbol == 777 ) continue; /* skip if not in Cond file */
      huff_encode(huff, symbol);
   }
      arc_put2(2);

    return 0;
}

void arc_put2 (unsigned bit)
{
 if ( bit == 1 ) put1flag = 0;/* start of symbol */
 else  /* closing file */
  if ( extraout == 1 ) outbit.wz(-1); /* an extra bit out */
   else outbit.wz(-2);
}
void arc_put1 (unsigned bit)
{
 if ( put1flag == 777 ) return;
 outbit.wz(bit);
 if (put1flag == 0 ){
  put1flag = 1;
  if ( bit == 0 ) extraout = 1;
 } else 
    if ( bit == 1 ) extraout = 0; 
}


unsigned arc_get2 ()
{
 get1flag = 0;/* start of symbol */
}
unsigned arc_get1 ()
{
  int x;
  x = inbit.r();
  if ( x < 0 ) {
//exit(3);
     if ( extrain == 1 && get1flag == 0 ) extrain = 300; /* drop it */
     else if ( extrain != 300 ) extrain = 400; /* last symbol processing */
     if ( x == -2 ) return 0; /* trailing zeros */
     else return 1;
  }
  if ( get1flag == 0 ) {
     if (x == 0 ) extrain = 1;
     get1flag = 1;
  } else 
     if (x == 1) extrain = 0; 
  return (unsigned int)x;
}
#endif
